﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Documents;
using System.Windows.Ink;
using System.Windows.Input;
using System.Windows.Media;
using System.Windows.Media.Animation;
using System.Windows.Shapes;
using System.Collections.Generic;
using System.Diagnostics;

namespace SilverlightImageBlur {
	public class PixelSnapBehavior {

		// Define the attached DependencyProperty

		public static DependencyProperty PixelSnapProperty = DependencyProperty.RegisterAttached(
			 "PixelSnap", typeof(PixelSnapType), typeof(PixelSnapBehavior), new PropertyMetadata(SnapPropertyChanged));

		// In the property changed notification method, we will add the element to a list of objects that will
		// be snapped when we get a LayoutUpdated event. We have to do a bunch of fancy stuff with weak references,
		// our own list, etc. because there is no Unloaded event.

		public static void SnapPropertyChanged(DependencyObject d, DependencyPropertyChangedEventArgs e) {
			PixelSnapType newSnap = (PixelSnapType)e.NewValue;
			int index = 0;

			while (index < _objects.Count) {
				if (_objects[index].Target == d)
					break;
				++index;
			}

			if (index < _objects.Count) {
				if (newSnap == PixelSnapType.None) {
					if (_objects[index].IsAlive) {
						Debug.WriteLine("Removing RenderTransform");
						((FrameworkElement)_objects[index].Target).RenderTransform = null;
					}
					_objects.RemoveAt(index);
				}
			}

			else if (newSnap != PixelSnapType.None) {
				_objects.Add(new WeakReference(d));
			}

			if (!_attached && _objects.Count > 0) {
				FrameworkElement element = d as FrameworkElement;
				if (element != null) {
					element.LayoutUpdated += new EventHandler(LayoutUpdated);
					_attached = true;
				}
			}
			else if (_attached && _objects.Count == 0) {
				FrameworkElement element = d as FrameworkElement;
				if (element != null) {
					element.LayoutUpdated -= new EventHandler(LayoutUpdated);
					_attached = false;
				}
			}
		}

		// The attached DependencyProperty setter
		public static void SetPixelSnap(DependencyObject obj, PixelSnapType value) {
			obj.SetValue(PixelSnapProperty, value);
		}

		// The attached DependencyProperty getter
		public static PixelSnapType GetPixelSnap(DependencyObject obj) {
			object result = obj.GetValue(PixelSnapProperty);
			return result == null ? PixelSnapType.None : (PixelSnapType)result;
		}



		// A utility method to remove all snapped objects from the list.

		public static void RemoveAll() {
			while (_objects.Count > 0) {
				if (_objects[0].IsAlive) {
					SetPixelSnap((DependencyObject)_objects[0].Target, PixelSnapType.None);
				}
				else {
					_objects.RemoveAt(0);
				}
			}
		}

		// The event handler for the LayoutUpdated event. It will snap everything that it thinks is
		// still alive. This is not 100% bulletproof but it should work in most scenarios.
		private static void LayoutUpdated(object sender, EventArgs e) {
			int index = 0;
			while (index < _objects.Count) {
				if (_objects[index].IsAlive == false) {
					_objects.RemoveAt(index);
				}

				else {
					Snap(_objects[index].Target as FrameworkElement);
					++index;
				}
			}
		}

		// Try to align an element on an integer pixel
		private static void Snap(FrameworkElement target) {
			if (target == null)
				return;
			PixelSnapType snap = PixelSnapBehavior.GetPixelSnap(target);
			// Remove existing transform
			TranslateTransform savedTransform = target.RenderTransform as TranslateTransform;
			if (savedTransform != null) {
				target.RenderTransform = null;
			}
			// Calculate actual location
			MatrixTransform globalTransform = target.TransformToVisual(Application.Current.RootVisual) as MatrixTransform;
			Point p = globalTransform.Matrix.Transform(_zero);
			double deltaX = snap == PixelSnapType.Closest ? Math.Round(p.X) - p.X : (int)p.X - p.X;
			double deltaY = snap == PixelSnapType.Closest ? Math.Round(p.Y) - p.Y : (int)p.Y - p.Y;
			// Set new transform
			if (deltaX != 0 || deltaY != 0) {
				if (savedTransform == null)
					savedTransform = new TranslateTransform();
				target.RenderTransform = savedTransform;
				savedTransform.X = deltaX;
				savedTransform.Y = deltaY;
			}
		}

		private static readonly Point _zero = new Point(0, 0);
		private static List<WeakReference> _objects = new List<WeakReference>();
		private static bool _attached = false;
	}
	public enum PixelSnapType {
		None,
		Closest,
		TopLeft
	}
}
